"""
Spinner
=======

#. :class:`.Spinner`

Generator of host guest conformations using nonbonded interactions.

"""

import numpy as np

import random

from .supramolecule import SupraMolecule
from .utilities import rotation_matrix_arbitrary_axis
from .potential import SpdPotential


class Spinner:
    """
    Generate host-guest conformations by rotating guest.

    A Metroplis MC algorithm is applied to perform rigid
    translations and rotations of the guest, relative to the host.

    """

    def __init__(
        self,
        step_size,
        rotation_step_size,
        num_conformers,
        max_attempts=1000,
        potential_function=SpdPotential(5),
        beta=2,
        random_seed=1000,
    ):
        """
        Initialize a :class:`Spinner` instance.

        Parameters
        ----------
        step_size : :class:`float`
            The relative size of the step to take during step.

        rotation_step_size : :class:`float`
            The relative size of the rotation to take during step.

        num_conformers : :class:`int`
            Number of conformers to extract.

        max_attempts : :class:`int`
            Maximum number of MC moves to try to generate conformers.

        potential_function : :class:`spd.Potential`
            Function to calculate potential energy of a
            :class:`spd.Supramolecule`

        beta : :class:`float`, optional
            Value of beta used in the in MC moves. Beta takes the
            place of the inverse boltzmann temperature.
            Defaults to 2.

        random_seed : :class:`int` or :class:`NoneType`, optional
            Random seed to use for MC algorithm. Should only be set to
            ``None`` if system-based random seed is desired. Defaults
            to 1000.

        """

        self._step_size = step_size
        self._num_conformers = num_conformers
        self._rotation_step_size = rotation_step_size
        self._max_attempts = max_attempts
        self._potential_function = potential_function
        self._beta = beta
        if random_seed is None:
            np.random.seed()
            random.seed()
        else:
            np.random.seed(random_seed)
            random.seed(random_seed)

    def _compute_potential(self, supramolecule):
        return self._potential_function.compute_potential(
            supramolecule
        )

    def _translate_atoms_along_vector(self, mol, vector):
        return mol.with_displacement(vector)

    def _rotate_atoms_by_angle(self, mol, angle, axis, origin):
        new_position_matrix = mol.get_position_matrix()
        # Set the origin of the rotation to "origin".
        new_position_matrix = new_position_matrix - origin
        # Perform rotation.
        rot_mat = rotation_matrix_arbitrary_axis(angle, axis)
        # Apply the rotation matrix on the position matrix, to get the
        # new position matrix.
        new_position_matrix = (rot_mat @ new_position_matrix.T).T
        # Return the centroid of the molecule to the original position.
        new_position_matrix = new_position_matrix + origin

        mol = mol.with_position_matrix(new_position_matrix)
        return mol

    def _test_move(self, curr_pot, new_pot):

        if new_pot < curr_pot:
            return True
        else:
            exp_term = np.exp(-self._beta*(new_pot-curr_pot))
            rand_number = random.random()

            if exp_term > rand_number:
                return True
            else:
                return False

    def _run_step(self, supramolecule, movable_components):

        component_list = list(supramolecule.get_components())
        component_sizes = {
            i: mol.get_num_atoms()
            for i, mol in enumerate(component_list)
        }
        max_size = max(component_sizes.values())

        # If movable components not set, select a guest randomly to
        # move and reorient.
        # Do not move or rotate largest component.
        if movable_components is None:
            # If there are different sizes.
            if len(set(component_sizes.values())) > 1:
                movable_components = tuple(
                    i for i in range(len(component_list))
                    if component_sizes[i] != max_size
                )
            # Else capture all!
            else:
                movable_components = tuple(
                    i for i in range(len(component_list))
                )

        targ_comp_id = random.choice([
            i for i in range(len(component_list))
            if i in movable_components
        ])

        targ_comp = component_list[targ_comp_id]

        # Random number from -1 to 1 for multiplying translation.
        rand = (random.random() - 0.5) * 2

        # Random translation direction.
        rand_vector = np.random.rand(3)
        rand_vector = rand_vector / np.linalg.norm(rand_vector)

        # Perform translation.
        translation_vector = rand_vector * self._step_size * rand
        targ_comp = self._translate_atoms_along_vector(
            mol=targ_comp,
            vector=translation_vector,
        )

        # Define a random rotation of the guest.
        # Random number from -1 to 1 for multiplying rotation.
        rand = (random.random() - 0.5) * 2
        rotation_angle = self._rotation_step_size * rand
        rand_axis = np.random.rand(3)
        rand_axis = rand_axis / np.linalg.norm(rand_vector)

        # Perform rotation.
        targ_comp = self._rotate_atoms_by_angle(
            mol=targ_comp,
            angle=rotation_angle,
            axis=rand_axis,
            origin=targ_comp.get_centroid(),
        )

        component_list[targ_comp_id] = targ_comp
        supramolecule = SupraMolecule.init_from_components(
            components=component_list,
        )

        nonbonded_potential = self._compute_potential(supramolecule)
        return supramolecule, nonbonded_potential

    def get_conformers(
        self,
        supramolecule,
        movable_components=None,
        verbose=False,
    ):
        """
        Get conformers of supramolecule.

        Parameters
        ----------
        supramolecule : :class:`.SupraMolecule`
            The supramolecule to optimize.

        movable_components : :class:`iterable` of :class:`int`,
        optional
            Components of supramolecule to move during simulation.

        verbose : :class:`bool`
            `True` to print some extra information.

        Yields
        ------
        conformer : :class:`.SupraMolecule`
            The host-guest supramolecule.

        """

        cid = 0
        nonbonded_potential = self._compute_potential(supramolecule)

        yield SupraMolecule(
            atoms=supramolecule.get_atoms(),
            bonds=supramolecule.get_bonds(),
            position_matrix=supramolecule.get_position_matrix(),
            cid=cid,
            potential=nonbonded_potential,
        )
        cids_passed = []
        for step in range(1, self._max_attempts):
            n_supramolecule, n_nonbonded_potential = self._run_step(
                supramolecule=supramolecule,
                movable_components=movable_components,
            )
            passed = self._test_move(
                curr_pot=nonbonded_potential,
                new_pot=n_nonbonded_potential
            )
            if passed:
                cid += 1
                cids_passed.append(cid)
                nonbonded_potential = n_nonbonded_potential
                supramolecule = SupraMolecule(
                    atoms=supramolecule.get_atoms(),
                    bonds=supramolecule.get_bonds(),
                    position_matrix=(
                        n_supramolecule.get_position_matrix()
                    ),
                    cid=cid,
                    potential=nonbonded_potential,
                )

                yield supramolecule

            if len(cids_passed) == self._num_conformers:
                break

        if verbose:
            print(
                f'{len(cids_passed)} conformers generated in {step} '
                'steps.'
            )

    def get_final_conformer(
        self,
        supramolecule,
        movable_components=None,
    ):
        """
        Get final conformer of supramolecule.

        Parameters
        ----------
        supramolecule : :class:`.SupraMolecule`
            The supramolecule to optimize.

        movable_components : :class:`iterable` of :class:`int`,
        optional
            Components of supramolecule to move during simulation.
            If `None`, then moved components are selected randomly,
            and the largest component (host) is not moved.

        Returns
        -------
        conformer : :class:`.SupraMolecule`
            The host-guest supramolecule.

        """

        for conformer in self.get_conformers(
            supramolecule=supramolecule,
            movable_components=movable_components,
        ):
            continue

        return conformer



import pytest
import numpy as np
import spindry as spd
import stk


@pytest.fixture(
    params=(
        (spd.Atom(id=0, element_string='N'), 0, 'N', 1.4882656711484),
        (spd.Atom(id=65, element_string='P'), 65, 'P', 1.925513788),
        (spd.Atom(id=2, element_string='C'), 2, 'C', 1.60775485914852),
    )
)
def atom_info(request):
    return request.param


@pytest.fixture(
    params=(
        (spd.Bond(id=0, atom_ids=(0, 1)), 0, 0, 1),
        (spd.Bond(id=65, atom_ids=(2, 3)), 65, 2, 3),
        (spd.Bond(id=2, atom_ids=(3, 4)), 2, 3, 4),
        (spd.Bond(id=3, atom_ids=(0, 9)), 3, 0, 9),
    )
)
def bond_info(request):
    return request.param


@pytest.fixture
def atoms():
    return [spd.Atom(0, 'C'), spd.Atom(1, 'C')]


@pytest.fixture
def position_matrix():
    return np.array([[0, 0, 0], [0, 1.5, 0]])


@pytest.fixture
def position_matrix2():
    return np.array([[0, 0, 0], [0, 3, 0]])


@pytest.fixture
def displacement():
    return np.array([0, 1, 0])


@pytest.fixture
def displaced_position_matrix():
    return np.array([[0, 1, 0], [0, 2.5, 0]])


@pytest.fixture
def centroid():
    return np.array([0, 0.75, 0])


@pytest.fixture
def num_atoms():
    return 2


@pytest.fixture
def molecule(atoms, position_matrix):
    return spd.Molecule(
        atoms=atoms,
        bonds=[],
        position_matrix=position_matrix
    )


@pytest.fixture
def components(atoms, position_matrix):
    return (
        spd.Molecule(
            atoms=(atoms[0], ),
            bonds=[],
            position_matrix=position_matrix[0]
        ),
        spd.Molecule(
            atoms=(atoms[1], ),
            bonds=[],
            position_matrix=position_matrix[1]
        ),
    )


@pytest.fixture
def smolecule(atoms, position_matrix):
    return spd.SupraMolecule(
        atoms=atoms,
        bonds=[],
        position_matrix=position_matrix,
    )


@pytest.fixture
def distances():
    return np.array([1, 1.2, 1.4, 1.6, 1.8, 2.0, 4])


@pytest.fixture
def nonbond_potentials():
    return [
        29.650582241279984, 0.0, -1.1965106134640058,
        -0.7315108180046076, -0.40042074284821816,
        -0.22239608831999993, -0.0036423427949999992
    ]


@pytest.fixture
def nb_mins():
    return [1, 2, 4, 3]


@pytest.fixture
def final_potential():
    return -0.04786452927586137


@pytest.fixture
def final_pos_mat():
    return np.array([
        [-0.19613681, -0.52185641, -0.26907745],
        [ 1.31683193,  2.57903968,  0.21028248]
    ])


@pytest.fixture
def spinner():
    return spd.Spinner(
        step_size=0.5,
        rotation_step_size=5,
        num_conformers=50,
    )


@pytest.fixture
def spdpotential():
    return spd.SpdPotential()


@pytest.fixture
def radii_combinations():
    return [
        (1.24235230881914, 1.24235230881914, 1.24235230881914),
        (1.24235230881914, 2.056812828, 1.64958256840957),
        (1.75038091053539, 1.24235230881914, 1.496366609677265),
        (1.60775485914852, 1.4882656711484, 1.54801026514846),
    ]


@pytest.fixture
def three_position_matrix():
    return np.array([
        [0, 0, 0],
        [0, 1.5, 0],
        [0, 3.0, 0]
    ])


@pytest.fixture
def smolecule_components(three_position_matrix):
    return spd.SupraMolecule(
        atoms=[spd.Atom(0, 'C'), spd.Atom(1, 'C'), spd.Atom(2, 'C')],
        bonds=[spd.Bond(0, (0, 1))],
        position_matrix=three_position_matrix,
    )


@pytest.fixture
def movable_components1():
    return (1, )


@pytest.fixture
def final_comp_pos_mat1():
    return np.array([
        [0, 0, 0],
        [0., 1.5, 0.],
        [0.59476412,  4.20474898, -1.24371808],
    ])


@pytest.fixture
def movable_components2():
    return (0, )


@pytest.fixture
def final_comp_pos_mat2():
    return np.array([
        [1.04061245,  0.44756825, -0.78994905],
        [-0.04983144,  1.15362244, -0.04000266],
        [0, 3.0, 0],
    ])



@pytest.fixture
def stk_spinner():
    return spd.Spinner(
        step_size=0.0,
        rotation_step_size=0.0,
        num_conformers=1,
    )


@pytest.fixture
def stk_host():
    # Building a cage from the examples on the stk docs.
    bb1 = stk.BuildingBlock(
        smiles='O=CC(C=O)C=O',
        functional_groups=[stk.AldehydeFactory()],
    )
    bb2 = stk.BuildingBlock('NCCN', [stk.PrimaryAminoFactory()])

    return stk.ConstructedMolecule(
        topology_graph=stk.cage.FourPlusSix(
            building_blocks=(bb1, bb2),
        ),
    )


@pytest.fixture
def stk_guest():
    return stk.BuildingBlock('C1CCCCC1')


@pytest.fixture
def stk_supramolecule(stk_host, stk_guest):
    host_guest = stk.ConstructedMolecule(
        topology_graph=stk.host_guest.Complex(
            host=stk.BuildingBlock.init_from_molecule(stk_host),
            guests=[stk.host_guest.Guest(stk_guest)],
        )
    )
    return spd.SupraMolecule(
        atoms=(
            spd.Atom(
                id=atom.get_id(),
                element_string=atom.__class__.__name__,
            ) for atom in host_guest.get_atoms()
        ),
        bonds=(
            spd.Bond(
                id=i,
                atom_ids=(
                    bond.get_atom1().get_id(),
                    bond.get_atom2().get_id(),
                )
            ) for i, bond in enumerate(host_guest.get_bonds())
        ),
        position_matrix=host_guest.get_position_matrix(),
    )


@pytest.fixture
def stk_supramolecule_by_comp(stk_host, stk_guest):

    host_molecule = spd.Molecule(
        atoms=(
            spd.Atom(
                id=atom.get_id(),
                element_string=atom.__class__.__name__,
            ) for atom in stk_host.get_atoms()
        ),
        bonds=(
            spd.Bond(
                id=i,
                atom_ids=(
                    bond.get_atom1().get_id(),
                    bond.get_atom2().get_id(),
                )
            ) for i, bond in enumerate(
                stk_host.get_bonds()
            )
        ),
        position_matrix=(
            stk_host.get_position_matrix()
        ),
    )
    guest_molecule = spd.Molecule(
        atoms=(
            spd.Atom(
                id=atom.get_id(),
                element_string=atom.__class__.__name__,
            ) for atom in stk_guest.get_atoms()
        ),
        bonds=(
            spd.Bond(
                id=i,
                atom_ids=(
                    bond.get_atom1().get_id(),
                    bond.get_atom2().get_id(),
                )
            ) for i, bond in enumerate(stk_guest.get_bonds())
        ),
        position_matrix=stk_guest.get_position_matrix(),
    )

    return spd.SupraMolecule.init_from_components(
        components=(host_molecule, guest_molecule),
    )



import numpy as np
from ..utilities import (
    is_equivalent_stk_molecule, is_equivalent_spd_molecule
)


def test_opt(spinner, smolecule, final_pos_mat, final_potential):
    test = spinner.get_final_conformer(smolecule)
    print(test.get_position_matrix())
    assert np.all(np.allclose(
        final_pos_mat,
        test.get_position_matrix(),
    ))
    assert np.isclose(
        spinner._compute_potential(test),
        final_potential
    )


def test_opt_stk(
    stk_spinner,
    stk_host,
    stk_guest,
    stk_supramolecule,
):
    test = stk_spinner.get_final_conformer(stk_supramolecule)
    is_equivalent_spd_molecule(test, stk_supramolecule)
    for i, comp in enumerate(list(test.get_components())):
        if i == 0:
            test_host = stk_host.with_position_matrix(
                comp.get_position_matrix()
            )
            is_equivalent_stk_molecule(test_host, stk_host)
        elif i == 1:
            test_guest = stk_guest.with_position_matrix(
                comp.get_position_matrix()
            )
            is_equivalent_stk_molecule(test_guest, stk_guest)


def test_opt_stk_components(
    stk_spinner,
    stk_host,
    stk_guest,
    stk_supramolecule,
):
    test = stk_spinner.get_final_conformer(stk_supramolecule)
    is_equivalent_spd_molecule(test, stk_supramolecule)
    for i, comp in enumerate(list(test.get_components())):
        if i == 0:
            test_host = stk_host.with_position_matrix(
                comp.get_position_matrix()
            )
            is_equivalent_stk_molecule(test_host, stk_host)
        elif i == 1:
            test_guest = stk_guest.with_position_matrix(
                comp.get_position_matrix()
            )
            is_equivalent_stk_molecule(test_guest, stk_guest)


def test_opt_test_move(spinner):
    # Do not test random component.
    assert spinner._test_move(curr_pot=-1, new_pot=-2)


def test_opt_setcomp1(
    spinner,
    smolecule_components,
    final_comp_pos_mat1,
    movable_components1,
):
    test = spinner.get_final_conformer(
        smolecule_components, movable_components1,
    )
    print(test.get_position_matrix())
    assert np.all(np.allclose(
        final_comp_pos_mat1,
        test.get_position_matrix(),
    ))


def test_opt_setcomp2(
    spinner,
    smolecule_components,
    final_comp_pos_mat2,
    movable_components2,
):
    test = spinner.get_final_conformer(
        smolecule_components, movable_components2,
    )
    print(test.get_position_matrix())
    assert np.all(np.allclose(
        final_comp_pos_mat2,
        test.get_position_matrix(),
    ))



import numpy as np


def test_smolecule_define_components(smolecule, components):
    tests = list(smolecule.get_components())

    assert len(tests) == len(components)
    for test, comp in zip(tests, components):
        for a1, a2 in zip(test.get_atoms(), comp.get_atoms()):
            assert a1.get_id() == a2.get_id()
            assert a1.get_element_string() == a2.get_element_string()

        assert np.all(np.allclose(
            test.get_position_matrix(),
            comp.get_position_matrix(),
        ))



import numpy as np


def test_nonbond_potential(
    spdpotential, distances, nonbond_potentials, nb_mins,
):

    for i, d in enumerate(distances):
        test = spdpotential._nonbond_potential(
            distance=d,
            sigmas=np.array(1.2),
        )
        assert np.isclose(test, nonbond_potentials[i], atol=1E-5)

    for i, _sigma in enumerate([1.2, 1.4, 1.8, 1.6]):
        test = spdpotential._nonbond_potential(
            distance=distances,
            sigmas=np.array(_sigma),
        )
        assert np.argmin(test**2) == nb_mins[i]


def test_combinations(spdpotential, radii_combinations):
    for combo in radii_combinations:
        test = spdpotential._mixing_function(combo[0], combo[1])

        assert test == combo[2]



import setuptools

setuptools.setup(
    name="SpinDry",
    version="0.0.8",
    author="Andrew Tarzia",
    author_email="andrew.tarzia@gmail.com",
    description="Contains MC algorithm for generating host-guest conformers.",
    url="https://github.com/andrewtarzia/SpinDry",
    packages=setuptools.find_packages(),
    install_requires=(
        'scipy',
        'matplotlib',
        'networkx',
        'numpy',
    ),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



"""
Atom
====

#. :class:`.Atom`

Atom class.

"""

from .radii import get_radius


class Atom:
    """
    Atom.

    """

    def __init__(self, id, element_string):
        """
        Initialize a :class:`Atom` instance.

        Parameters
        ----------
        id : :class:`int`
            ID to be assigned to atom.

        element_string : :class:`str`
            Atom element symbol as string.

        """

        self._id = id
        self._element_string = element_string
        self._radius = get_radius(element_string)

    def get_id(self):
        """
        Get atom ID.

        """
        return self._id

    def get_element_string(self):
        """
        Get atom element symbol.

        """
        return self._element_string

    def get_radius(self):
        """
        Get atomic radius (STREUSEL).

        """

        return self._radius

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return f'{self.get_element_string()}(id={self.get_id()})'



from spindry import get_atom_distance
import numpy as np


def test_get_atom_distance():
    position_matrix = np.array([
        [0, 0, 0],
        [0, 1, 0],
        [2, 0, 0],
    ])
    assert get_atom_distance(position_matrix, 0, 1) == 1
    assert get_atom_distance(position_matrix, 0, 2) == 2
